#	  ****************************************
#	  *           unload_prop.pl             *
#	  * (c) Copyright 2009 Philip C. Pilgrim *
#	  *     See end of file for licensing.   *
#	  ****************************************
#	
#	The program unload_prop.pl connects to the Propeller, uploads a small EEPROM
#	dumping program to RAM, dumps the contents of EEPROM to the PC, and writes the
#	data to an .eeprom file.
#	
#	Here is the command line syntax to invoke it:
#	
#	 perl unload_prop.pl XINFREQ XTAL FILE
#	
#	All the parameters ore optional. Their valid and default values are shown below:
#	
#	 XINFREQ can be any number designating the crystal frequency, either in Hz or MHz.
#	   Default is 5_000_000 if omitted.
#	
#	 XTAL can be XTAL1, XTAL2, XTAL3, or XINPUT. Default is XTAL1 if omitted.
#	
#	 FILE is any valid file name. If the file name or directory path contains spaces,
#	   FILE must be enclosed in "quotes". If ".eeprom" is omitted from the end, it will
#	   be added. Default is "unload.eeprom" in the current directory if omitted.

use strict;

use Win32::SerialPort;
use File::Spec;
use sigtrap 'handler' => \&Interrupt, 'normal-signals';

#perl2exe_bundle "c:\Program Files\Parallax Inc\Propellent v1.2\propellent.exe"

$| = 1;
print "\n";
my $self = $0;
$self =~ s/\\/\//g;
$self =~ s/(.*)\/[^\/]*/$1/;
my $INIT_DIR = File::Spec->canonpath($self);
my $TEMP_DIR = ($ENV{TEMP} || $ENV{TMP} || $ENV{WINDIR} || '/tmp' ) . "/p2xtmp-$$";

$TEMP_DIR = $INIT_DIR unless -e "$TEMP_DIR/propellent.exe";

my $Port;
my $XTAL = 'XTAL1';
my $XIN = 5000000;
my $FILE = "$INIT_DIR\\unload.eeprom";
my $PLL;
my $BAUD;

while (@ARGV) {
	my $arg = shift(@ARGV);
	if ($arg =~ m/^(XTAL1|XTAL2|XTAL3|XINPUT)$/) {
		$XTAL = $1
	} elsif ($arg =~ m/^([\d_\.]+)$/) {
		$XIN = $1;
		$XIN =~ s/_//g;
		$XIN *= 1000000 if $XIN <= 200
	} elsif (@ARGV == 0) {
		$FILE = $arg;
		$FILE .= ".eeprom" unless $FILE =~ m/\.eeprom$/
	} else {
		Error("Unrecognized argument: $arg")
	}
}
Error("Can't open file: $FILE") unless open EEPROM, ">$FILE";
binmode EEPROM;

foreach my $pll (16, 8, 4, 2, 1) {
	if ($pll * $XIN <= 100000000) {
		$PLL = "PLL$pll" . 'X';
		$BAUD = $pll * $XIN >= 80000000 ? 19200 : 9600;
		last
	}
}

my $Spin = GetSpin();

open SPIN, ">$TEMP_DIR/unload.spin" or Error("Can't create \"unload.spin\" file");
print SPIN $Spin;
close SPIN;

my $port;

my $result = `"$TEMP_DIR/propellent.exe" "$TEMP_DIR/unload.spin"`;

if ($result && $result =~ m/ERR:\d+-([^\n\r]*)/) {
	Error($1)
} elsif ($result && $result =~ m/EVT:505[^\n\r]+(COM\d+)/i) {
	$port = $1;
	$Port = "\\\\.\\" . $1
} else {
	Error('Propellent failed to upload "unload.spin".')
}

print "  Using clock settings:\n\n    _clkmode     = $XTAL + $PLL\n    _xinfreq     = $XIN\n\n";
print "  Output file: $FILE.\n\n";
print "  Connecting to $port at $BAUD baud.\n\n";

unless ($Port = new Win32::SerialPort($Port)
	and	$Port->baudrate($BAUD)
	and $Port->parity('none')
	and $Port->databits(8)
	and $Port->stopbits(2)
	and $Port->handshake('none')
	and $Port->dtr_active('N')
	and $Port->buffers(16384,4096)
	and $Port->binary(1)
	and $Port->datatype('raw')
	and $Port->write_settings)
{
	Error("Can't open serial port to download data.")
}

sleep 2;
my $eedata = $Port->input;
Error("Download failed to start correctly or EEPROM header corrupt. Check clock settings.") unless ($eedata =~ m/\<\r\n?[0-9A-F]{12}1000/);

my $count = 32;
print "  Downloading $count";
my $plen = 0;

while (1) {
	my $inp = $Port->input;
	$eedata .= $inp;
	last if $inp =~ m/\>/;
	my $len = length($eedata);
	if ($len - $plen > 2048) {
		print chr(8), chr(8), sprintf('%2.1d', $count);
		$plen += 2048;
		$count--
	}	
}

$Port->close;

print chr(8), chr(8), "done.\n\n";

$eedata =~ s/[\r\n]//g;
$eedata =~ m/\<([A-F0-9]+)\>/ or Error('Data download error.');
length($eedata = $1) == 65536 or Error('Data download error.');
my @eedata = map {hex(substr($1, $_ * 2, 2))} (0 .. 32767);
my $chksum = 0;
foreach (@eedata) {
	$chksum += $_
}
if ($chksum &= 0xff) {
	print "  WARNING: Checksum is non-zero. EEPROM may be corrupt.\n  Checksum will be adjusted to make it valid.\n\n";
	$eedata[5] = ($eedata[5] - $chksum) & 0xff
}
print EEPROM chr($_) foreach (@eedata);
close EEPROM;
print "  EEPROM file has been written. Finished.\n\n";

Quit();

sub Interrupt {
	print "\n\n  *** Process interrupted. ***\n\n";
	Quit()
}

sub Error {
	undef $Port if $Port;
	print "\n  ERROR: $_[0]\n";
	Quit()
}

sub Quit {
	unlink "$TEMP_DIR/unload.spin" if -e "$TEMP_DIR/unload.spin";
	undef $Port if $Port;
	exit
}

sub GetSpin {
	return <<END_OF_SPIN;
CON

  _clkmode     = $XTAL + $PLL
  _xinfreq     = $XIN

  SCL          =  28
  SDA          =  29
  SOUT         =  30

VAR

  long  BaudClock

PUB start | i, data

  waitcnt(cnt + clkfreq)
  BaudClock := clkfreq / 19200
  outa[SOUT]~~
  dira[SOUT]~~
  repeat 8
    eestart
  i~
  repeat
    eestart
  until eewrbyte(%1010_0000) == 0 or ++i == 5
  if (i < 5)
    eewrbyte(0)
    eewrbyte(0)
    eestart
    eewrbyte(%1010_0001)
    serout("<")
    serout(13)
    repeat i from 0 to 32767
      serhex(eerdbyte(i <> 32767))
      if i & 31 == 31
        serout(13)
  serout(">")
  repeat

PUB eestart

  dira[SDA]~~
  outa[SCL]~~
  dira[SCL]~~
  dira[SDA]~
  dira[SDA]~~
  outa[SCL]~

PUB eewrbyte(data) | i

  data := data << 1 | 1
  repeat i from 0 to 8
    dira[SDA] := not ((data <<= 1) & \$200)
    outa[SCL]~~
    outa[SCL]~

PUB eerdbyte(ack) : data | i

  data~
  repeat i from 0 to 7
    outa[SCL]~~
    data := data << 1 | ina[SDA]
    outa[SCL]~
  if (ack)
    dira[SDA]~~
  outa[SCL]~~
  outa[SCL]~
  dira[SDA]~

PUB serhex(data)

  serout(lookupz((data >> 4) & \$F : "0".."9", "A".."F"))
  serout(lookupz(data & \$F : "0".."9", "A".."F"))

PUB serout(data) | time

  data := (data | \$300) << 2
  time := cnt
  repeat 11
    waitcnt(time += BaudClock)
    outa[SOUT] := (data >>= 1) & 1
END_OF_SPIN
}

#	LICENSE INFORMATION
#	    
#	This program is released for use under the terms of the Perl Artistic License,
#	which may be downloaded here: http://dev.perl.org/licenses/artistic.html
#	
#	THIS PROGRAM IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#	IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#	WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  